# pylint: disable=no-name-in-module
import ctypes
import sys
import webbrowser
from datetime import datetime, timedelta

from PySide6.QtCore import Qt, QTimer, Slot
from PySide6.QtWidgets import (
    QApplication,
    QComboBox,
    QDialog,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
)

from neverlate.constants import APP_NAME, OUTPUT_DISMISS, OUTPUT_SNOOZE
from neverlate.utils import get_icon, now_datetime, pretty_datetime

LOCAL_TIMEZONE = datetime.now().astimezone().tzinfo

RESTART_TIMER = 15 * 1000  # 15 seconds  (in milliseconds)

SNOOZE_UNTIL_MINUTES = [0, 0.1, 1, 3, 5, 10, 15, 30]  # Minute for the snooze until...
FONT_SIZE_MULTIPLIER = 1  # Multiply font size


class AlertDialog(QDialog):
    """Pop up alert dialog."""

    has_shown_maximized: bool = False

    def __init__(self, title: str, start_time: str, video_uri: str) -> None:
        super().__init__()
        self.start_time = datetime.fromisoformat(start_time)
        self.video_uri = video_uri
        self.setWindowTitle(
            f"Don't Be Late: {title} @ {pretty_datetime(self.start_time)}"
        )
        self.setWindowIcon(get_icon("tray_icon.png"))

        self.start_time_label = QLabel("Starts at: " + pretty_datetime(self.start_time))
        self.start_time_label.setAlignment(Qt.AlignCenter)
        self.time_to_event_label = QLabel(" ")
        self.standard_font_size = self.time_to_event_label.font().pointSize()
        self.time_to_event_label.setAlignment(Qt.AlignCenter)

        self.button_accept = QPushButton("Dismiss")
        self.button_accept.clicked.connect(self.dismiss)
        self.button_join = QPushButton("Joing Meeting")
        if "meet.google" in self.video_uri:
            self.button_join.setText("Join Meeting (Google Meet)")
        self.button_join.clicked.connect(self.dismiss_and_join)
        self.button_join.setIcon(get_icon("video.png"))
        if not self.video_uri:
            self.button_join.setVisible(False)

        self.snooze_until_combo_box = QComboBox()
        self.snooze_until_combo_box.addItems(["Snooze Until..."])
        self.snooze_until_combo_box.currentIndexChanged.connect(
            self.snooze_until_combo_box_changed
        )
        self.snooze_for_combo_box = QComboBox()
        self.snooze_for_combo_box.addItems(
            ["Snooze for...", "1 min", "3 min", "5 min", "10 min", "15 min", "30 min"]
        )
        self.snooze_for_combo_box.currentTextChanged.connect(
            self.snooze_for_combo_box_changed
        )

        self.update_timer = QTimer()
        self.update_timer.timeout.connect(self.update_ui)
        self.update_timer.start(0.9 * 1000)  # 1 sec intervall

        # Auto close the dialog and re-open after X seconds. This makes sure the dialog wasn't opened on a different
        # workspace than the user is currently on.
        # TODO: do NOT
        self.kill_timer = QTimer()
        self.kill_timer.timeout.connect(self.kill_timer_complete)
        self.kill_timer.setSingleShot(True)
        self.kill_timer.start(RESTART_TIMER)

        self.enable_widgets_timer = QTimer()
        self.enable_widgets_timer.timeout.connect(self.enable_widgets)
        self.enable_widgets_timer.start(0.75 * 1000)
        self.enable_widgets(value=False)

        # self.setWindowFlags(Qt.FramelessWindowHint)
        self.layout_ui(title)
        self.update_ui()

    def enable_widgets(self, value: bool = True):
        """Enable buttons.  Auto fired after the app is opened."""
        self.button_join.setEnabled(value)
        self.button_accept.setEnabled(value)

    def layout_ui(self, title: str):
        """Define the main UI layout."""
        main_layout = QVBoxLayout()
        self.setLayout(main_layout)

        main_layout.addStretch()
        main_label = QLabel(title)
        main_label.setAlignment(Qt.AlignCenter)
        font = main_label.font()
        font.setPointSize(self.standard_font_size * 3 * FONT_SIZE_MULTIPLIER)
        font.setBold(True)
        main_label.setFont(font)
        main_layout.addWidget(main_label)
        main_layout.addWidget(self.start_time_label)
        main_layout.addWidget(self.time_to_event_label)

        # Primary buttons
        buttons_layout = QHBoxLayout()
        button_t3 = QPushButton("Snooze for 3 minutes")
        button_t3.clicked.connect(lambda: self.snooze(3))
        button_t1 = QPushButton("Snooze for 1 minute")
        button_t1.clicked.connect(lambda: self.snooze(1))
        buttons_layout.addStretch()
        # buttons_layout.addWidget(button_t3)
        # buttons_layout.addWidget(button_t1)
        buttons_layout.addWidget(self.snooze_until_combo_box)
        buttons_layout.addWidget(self.snooze_for_combo_box)
        buttons_layout.addSpacing(50)
        # buttons_layout.addStretch()
        buttons_layout.addWidget(self.button_join)
        buttons_layout.addWidget(self.button_accept)
        buttons_layout.addStretch()

        main_layout.addSpacing(15)
        main_layout.addLayout(buttons_layout)
        main_layout.addStretch()
        self.setWindowFlags(Qt.WindowStaysOnTopHint)

    def close(self) -> bool:
        # return super().close()
        self.snooze(0)

    def dismiss_and_join(self):
        """Dismiss teh alert, and open the video URI link in a web browser."""
        webbrowser.open(self.video_uri, autoraise=True)
        self.dismiss()

    def dismiss(self):
        """Dismiss the alert permanently."""
        print(OUTPUT_DISMISS)
        sys.exit(0)

    def kill_timer_complete(self):
        """Called when the restart timer ends. If the UI doesn't have focus, restart."""
        # TODO: check if the mouse is over the dialog
        # print(OUTPUT_RESTART)
        self.snooze(0.001)
        # sys.exit(0)

    def update_snooze_for_combo_box_icons(self):
        """Add a warning icon to any combo box item that would make the user late to the meeting."""
        now = now_datetime()
        for idx in range(1, self.snooze_for_combo_box.count()):
            minutes = int(self.snooze_for_combo_box.itemText(idx).split()[0])
            if now + timedelta(minutes=minutes) > self.start_time:
                self.snooze_for_combo_box.setItemIcon(idx, get_icon("warning.png"))

    def update_snooze_until_options(self):
        """Update the valid options for the snooze until"""
        target_items = ["Snooze Until..."]
        current_count = self.snooze_until_combo_box.count()
        for minutes_before in SNOOZE_UNTIL_MINUTES[1:]:
            next_alert_time = self.start_time - timedelta(minutes=int(minutes_before))
            if next_alert_time < now_datetime():
                break
            label = pretty_datetime(next_alert_time)
            if int(minutes_before):
                label += f" ({int(minutes_before)} min before event)"
            target_items.append(label)
        if len(target_items) > current_count:
            self.snooze_until_combo_box.addItems(target_items[1:])
        else:
            while len(target_items) < self.snooze_until_combo_box.count():
                self.snooze_until_combo_box.removeItem(
                    self.snooze_until_combo_box.count() - 1
                )

    def update_ui(self):
        """Update the UI."""
        # If the user is observing the dialog, restart the kill timer
        if self.underMouse():
            self.kill_timer.start(RESTART_TIMER)
        self.update_snooze_for_combo_box_icons()
        self.update_snooze_until_options()
        # self.setWindowTitle(f"Time to event: {self.kill_timer.remainingTime()}")

        now = datetime.now(LOCAL_TIMEZONE)
        seconds_till = (self.start_time - now).total_seconds()
        min_, sec = divmod(int(abs(seconds_till)), 60)
        sec = str(sec).rjust(2, "0")
        hrs, min_ = divmod(min_, 60)
        hrs = f"{hrs}:" if hrs else ""
        min_ = str(min_).rjust(2, "0") if hrs else str(min_)
        if seconds_till <= 0:
            # Late
            self.setWindowIcon(get_icon("warning.png"))
            self.time_to_event_label.setText(
                f"YOU'RE LATE! (How could this happen!?)\nTHE EVENT STARTED {hrs}{min_}:{sec} AGO!!"
            )
            font = self.time_to_event_label.font()
            font.setBold(True)
            font.setPointSize(self.standard_font_size * 2 * FONT_SIZE_MULTIPLIER)
            self.time_to_event_label.setFont(font)
            self.setStyleSheet("background-color: rgb(255, 55, 55);")
            if not self.has_shown_maximized:
                self.showMaximized()
                self.has_shown_maximized = True
            self.snooze_until_combo_box.setVisible(False)
        else:
            if seconds_till < 60 * 2:
                # Warning!
                self.setStyleSheet("background-color: rgb(255, 245, 55);")

                font = self.time_to_event_label.font()
                font.setBold(True)
                font.setPointSize(15)
                self.time_to_event_label.setFont(font)
            # Plenty of time...
            self.time_to_event_label.setText(f"Time to event: {hrs}{min_}:{sec}")

    def snooze(self, minutes: float):
        """Close the dialog and snooze for X minutes."""
        print(f"{OUTPUT_SNOOZE} {int(minutes * 60)}")
        sys.exit(0)

    @Slot(str)
    def snooze_for_combo_box_changed(self, new_text: str):
        """Callback when the snooze for comobo box is changed."""
        minutes = int(new_text.split()[0])
        self.snooze(minutes)

    @Slot(int)
    def snooze_until_combo_box_changed(self, new_idx: int):
        """Callback when the snooze until combo box is changed."""
        minutes_before = SNOOZE_UNTIL_MINUTES[new_idx]
        next_alert_time = self.start_time - timedelta(minutes=minutes_before)
        time_to_snooze = now_datetime() - next_alert_time
        self.snooze(-time_to_snooze.total_seconds() / 60)


if __name__ == "__main__":
    # _, title, start_time_iso, video_uri = sys.argv
    # start_time = datetime.fromisoformat(start_time_iso)

    app = QApplication()
    if hasattr(ctypes, "windll"):
        # Rename the process so we can get a better icon for Windows
        myappid = f"bw.{APP_NAME.lower()}.dialog" + str(sys.argv[1:])
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

    app.setWindowIcon(get_icon("tray_icon.png"))  # Mac OSX
    d = AlertDialog(*sys.argv[1:])  # pylint: disable=no-value-for-parameter
    d.show()
    app.exec()



"""Main app entry point."""
from __future__ import annotations

import json
import os
from typing import Any, ClassVar

from neverlate.utils import app_local_data_dir


class _Preferences:

    _pref_file_path: ClassVar[str] = os.path.join(
        app_local_data_dir(), "preferences.json"
    )

    alert_padding: int  # Minutes before an event that an alert should be displayed
    calendar_visibility: dict[str, bool]  # Wether the calendar is enabled or not
    download_cal_freq: int  # Frequency in minutes that the calendar + events are downloaded

    def __init__(self) -> None:
        self.alert_padding = 5
        self.calendar_visibility = {}  # ty
        self.download_cal_freq = 5

    def deserialize(self, **kwargs: Any):
        """Load preferences from a dictionary."""
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def load(self):
        """Load preferences from disk."""
        if not os.path.exists(self._pref_file_path):
            return
        try:
            with open(self._pref_file_path, "r") as pref_file:
                data = json.load(pref_file)
        except:
            print("WARNING: unable to load prefernces!")
            data = {}
        self.deserialize(**data)

    def save(self):
        """Save the preferences to disk."""
        data = self.serialize()
        with open(self._pref_file_path, "w") as pref_file:
            json.dump(data, pref_file, sort_keys=True, indent=4)

    def serialize(self) -> dict[str, Any]:
        """Save preferences to a dictionary."""
        return {
            "alert_padding": self.alert_padding,
            "calendar_visibility": self.calendar_visibility,
            "download_cal_freq": self.download_cal_freq,
        }


PREFERENCES = _Preferences()
PREFERENCES.load()



from __future__ import annotations

import typing
import webbrowser
from datetime import timedelta

from PySide6.QtCore import QTimer  # QThreadPool
from PySide6.QtCore import QRect, Qt, QThread, Signal, Slot
from PySide6.QtGui import QAction, QColor, QCursor, QDesktopServices, QFont, QWindow
from PySide6.QtWidgets import (  # pylint: disable=no-name-in-module
    QAbstractScrollArea,
    QDialog,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QMenu,
    QPushButton,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QWidget,
)

from neverlate.utils import get_icon, now_datetime, pretty_datetime

if typing.TYPE_CHECKING:
    from PySide6.QtCore import QEvent

    from neverlate.event_alerter import EventAlerter

# Table columns
TABLE_SUMMARY = 0
TABLE_TIME_TILL_ALERT = 1
TABLE_EVENT_TIMES = 2
TABLE_CALENDAR = 3


# Colors
BLACK = QColor(*(3 * [0]))
DARK_GREY = QColor(*(3 * [50]))
LIGHT_GREY = QColor(*(3 * [200]))
LIGHT_GREEN = QColor(200, 255, 200)
LIGHT_RED = QColor(255, 200, 200)
LIGHT_YELLOW = QColor(255, 255, 200)
WHITE = QColor(*(3 * [255]))


class TimeTillAlertWidget(QWidget):
    """Widget to show some the time till an event (4:32) and a button to re-trigger teh alert."""

    def __init__(self, text: str):
        super().__init__()

        box_layout = QHBoxLayout()
        box_layout.setContentsMargins(0, 0, 0, 0)
        self.main_text = QLabel(text)
        self.reset_button = QPushButton(icon=get_icon("alarm.png"))
        box_layout.addWidget(self.main_text)
        box_layout.addWidget(self.reset_button)
        self.setLayout(box_layout)


class MainDialog(QDialog):
    """Main dialog to show general info."""

    def __init__(self) -> None:
        super().__init__()
        self.setWindowTitle("NeverLate")
        self.setWindowIcon(get_icon("tray_icon.png"))

        # Right click menu for the QTable
        self.table_menu = QMenu(self)
        self.label_section = self.table_menu.addAction("<Event Summary>")
        self.table_menu.addSeparator()

        self.trigger_alert_action = self.table_menu.addAction("Reset/Retrigger Alert")
        self.trigger_alert_action.setIcon(get_icon("alarm.png"))
        self.join_meeting_action = self.table_menu.addAction("Join Meeting")
        self.join_meeting_action.setIcon(get_icon("video.png"))

        self.update_now_button = QPushButton("Update Now")
        self.time_to_update_label = QLabel()
        self.row_to_event_alerter = {}  # type: dict[int, EventAlerter]

        self.event_table = QTableWidget(0, 4)
        self.event_table.setHorizontalHeaderItem(
            TABLE_SUMMARY, QTableWidgetItem("Event")
        )
        self.event_table.setHorizontalHeaderItem(
            TABLE_EVENT_TIMES, QTableWidgetItem("Time")
        )
        self.event_table.setHorizontalHeaderItem(
            TABLE_TIME_TILL_ALERT, QTableWidgetItem("Tim Till Alert")
        )
        self.event_table.setHorizontalHeaderItem(
            TABLE_CALENDAR, QTableWidgetItem("Calendar")
        )
        # self.event_table.horizontalHeader().hide()
        self.event_table.verticalHeader().hide()
        self.event_table.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)
        self.event_table.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeToContents
        )
        self.event_table.horizontalHeader().setStretchLastSection(True)
        self.event_table.setEditTriggers(QTableWidget.NoEditTriggers)
        self.event_table.setSelectionMode(QTableWidget.NoSelection)
        self.event_table.setFocusPolicy(Qt.NoFocus)
        self.event_table.contextMenuEvent = self.table_context_menu
        # table->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.event_table)
        # main_layout.addStretch()

        layout = QHBoxLayout()
        layout.addStretch()
        layout.addWidget(self.time_to_update_label)
        layout.addWidget(self.update_now_button)

        main_layout.addLayout(layout)
        self.setLayout(main_layout)

    def join_meeting(self, alerter: EventAlerter):
        """Join a meeting - opening a browser to the given URL"""
        url = alerter.time_event.get_video_url()
        if url:
            webbrowser.open(url, autoraise=True)

    def table_context_menu(self, event: QEvent):
        """
        Display the right-click context menu - letting a user join a meeting directly, or re-trigger an alert.

        Args:
            event (QEvent): Event triggered
        """
        row = self.event_table.rowAt(event.pos().y())
        alerter = self.row_to_event_alerter.get(row)
        if alerter is None:
            # This should not happen...
            return

        self.label_section.setText(alerter.time_event.summary)

        # Trigger alert
        self.trigger_alert_action.setVisible(alerter.has_alerted)
        self.trigger_alert_action.triggered.connect(alerter.reset_alert)

        # Join meeting
        url = alerter.time_event.get_video_url()
        self.join_meeting_action.setVisible(bool(url))
        self.join_meeting_action.setText(
            "Join Meeting (Google Meet)" if "meet.google" in url else "Join Meeting"
        )
        self.join_meeting_action.triggered.connect(lambda: self.join_meeting(alerter))

        # Show the menu
        self.table_menu.popup(QCursor.pos())

    def update_table_with_events(self, alerters: list[EventAlerter]):
        """Update the table with the specified alerter events."""
        # TODO: break this function up
        orig_row_count = self.event_table.rowCount()
        self.event_table.setRowCount(len(alerters))
        alerters.sort(key=lambda a: a.time_event.start_time)
        now = now_datetime()
        self.row_to_event_alerter.clear()
        for idx, alerter in enumerate(alerters):
            self.row_to_event_alerter[idx] = alerter

            self.event_table.setItem(
                idx, TABLE_SUMMARY, QTableWidgetItem(alerter.time_event.summary)
            )

            # Start time
            start_label = pretty_datetime(alerter.time_event.start_time).split()[0]
            end_label = pretty_datetime(alerter.time_event.end_time)
            self.event_table.setItem(
                idx,
                TABLE_EVENT_TIMES,
                QTableWidgetItem(f"{start_label} - {end_label}"),
            )

            # Time till alert
            time_till_alert = alerter.time_till_alert()
            if time_till_alert <= 0:
                time_till_alert = "---"
            else:
                min_, secs = divmod(time_till_alert, 60)
                hours, min_ = divmod(min_, 60)
                if hours:
                    secs = str(secs).zfill(2)
                    min_ = str(min_).zfill(2)
                    time_till_alert = f"{hours}:{min_}:{secs}"
                else:
                    secs = str(secs).zfill(2)
                    time_till_alert = f"{min_}:{secs}"

                time_till_alert = str(time_till_alert)

            # Set the base text
            self.event_table.setItem(
                idx, TABLE_TIME_TILL_ALERT, QTableWidgetItem(time_till_alert)
            )

            # Calendar
            self.event_table.setItem(
                idx,
                TABLE_CALENDAR,
                QTableWidgetItem(alerter.time_event.calendar.summary),
            )

            # =================== Styles =====================
            if alerter.time_event.has_declined():
                # User declined. Set strikethrough font & italic
                # for column in range(self.event_table.columnCount()):
                item = self.event_table.item(idx, TABLE_SUMMARY)
                font = QFont()
                font.setItalic(True)
                font.setStrikeOut(True)
                item.setFont(font)
            if now > alerter.time_event.end_time:
                # Meeting is over. 'Disable' it.
                for column in range(self.event_table.columnCount()):
                    item = self.event_table.item(idx, column)
                    background = item.background()
                    background.setColor(LIGHT_GREY)
                    foreground = item.foreground()
                    foreground.setColor(DARK_GREY)
                    background.setStyle(Qt.BrushStyle.SolidPattern)
                    foreground.setStyle(Qt.BrushStyle.SolidPattern)
                    item.setBackground(background)
                    item.setForeground(foreground)
            elif alerter.time_event.start_time < now < alerter.time_event.end_time:
                # Meeting is happening
                for column in range(self.event_table.columnCount()):
                    if alerter.dismissed_alerts:  # User is in the meeting (in theory)
                        bg_color = LIGHT_GREEN  # Light green
                    else:  # User should be in th emeeting
                        bg_color = LIGHT_RED
                    item = self.event_table.item(idx, column)
                    background = item.background()
                    foreground = item.foreground()
                    foreground.setColor(BLACK)
                    background.setColor(bg_color)
                    background.setStyle(Qt.BrushStyle.SolidPattern)
                    foreground.setStyle(Qt.BrushStyle.SolidPattern)
                    item.setBackground(background)
                    item.setForeground(foreground)
                    font = item.font()
                    font.setBold(True)
                    item.setFont(font)
            elif now + timedelta(minutes=30) > alerter.time_event.start_time:
                # Meeting is coming up soon
                for column in range(self.event_table.columnCount()):
                    item = self.event_table.item(idx, column)
                    background = item.background()
                    background.setColor(LIGHT_YELLOW)
                    foreground = item.foreground()
                    foreground.setColor(BLACK)
                    background.setStyle(Qt.BrushStyle.SolidPattern)
                    foreground.setStyle(Qt.BrushStyle.SolidPattern)
                    item.setBackground(background)
                    item.setForeground(foreground)
                    font = item.font()
                    font.setBold(True)
                    item.setFont(font)
            else:
                # Future event
                for column in range(self.event_table.columnCount()):
                    item = self.event_table.item(idx, column)
                    background = item.background()
                    background.setColor(WHITE)
                    foreground = item.foreground()
                    foreground.setColor(BLACK)
                    background.setStyle(Qt.BrushStyle.SolidPattern)
                    foreground.setStyle(Qt.BrushStyle.SolidPattern)
                    item.setBackground(background)
                    item.setForeground(foreground)
                    font = item.font()
                    item.setFont(font)

        # self.event_table.resizeColumnsToContents()
        if orig_row_count != len(alerters):
            self.adjustSize()



"""Main app entry point."""
from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtWidgets import (
    QCheckBox,
    QDialog,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QScrollArea,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from neverlate.preferences import PREFERENCES
from neverlate.utils import get_icon

if TYPE_CHECKING:
    from neverlate.google_cal_downloader import Calendar

# pylint: disable=no-name-in-module


# TODO: implement


class PreferencesDialog(QDialog):  # pylint: disable=too-few-public-methods
    """Preferences dialog panes"""

    def __init__(self) -> None:
        super().__init__()
        self.setWindowTitle("Preferences")
        self.setWindowIcon(get_icon("tray_icon.png"))

        self.alert_padding_sb = QSpinBox()
        self.alert_padding_sb.setValue(PREFERENCES.alert_padding)

        self.download_cal_freq_sb = QSpinBox()
        self.download_cal_freq_sb.setValue(PREFERENCES.download_cal_freq)
        self.download_cal_freq_sb.setMinimum(3)

        # Log out button
        self.logout_button = QPushButton("Logout")
        self.logout_button.pressed.connect(self.close)

        # Close buttons
        self.apply_button = QPushButton("Apply")
        self.apply_button.pressed.connect(self.save)
        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.pressed.connect(self.close)
        # self.button.clicked.connect(close_s)
        self.scroll_widget = QWidget()
        self.calendars = []  # type: list[Calendar]
        self.calendar_toggles = {}  # type: dict[str, QCheckBox]

        self.layout_ui()

    def layout_ui(self):
        """Lays out all the UI elements"""
        main_layout = QVBoxLayout()

        toggle_layout = QFormLayout()
        toggle_layout.addRow(
            "Minutes before an event to show an alert dialog", self.alert_padding_sb
        )
        toggle_layout.addRow(
            "Frequency in minutes to check for new events", self.download_cal_freq_sb
        )
        main_layout.addLayout(toggle_layout)

        # Calendars
        main_layout.addWidget(QLabel("Calendar(s)"))
        scroll_area = QScrollArea()
        scroll_layout = QVBoxLayout()
        scroll_layout.setContentsMargins(0, 0, 0, 0)

        scroll_area.setWidget(self.scroll_widget)
        self.scroll_widget.setLayout(scroll_layout)
        scroll_area.setWidgetResizable(True)
        # scroll_area.setFixedHeight(200)
        # scroll_area.setMaximumHeight(600)
        scroll_area.setMinimumHeight(130)
        main_layout.addWidget(scroll_area)

        # Close/accept buttons
        button_box = QHBoxLayout()
        button_box.addWidget(self.logout_button)
        button_box.addStretch()
        button_box.addWidget(self.apply_button)
        button_box.addWidget(self.cancel_button)
        main_layout.addLayout(button_box)

        self.setLayout(main_layout)
        self.adjustSize()

    def save(self):
        """Save the preferences and close the dialog."""
        PREFERENCES.alert_padding = self.alert_padding_sb.value()
        PREFERENCES.download_cal_freq = self.download_cal_freq_sb.value()
        PREFERENCES.calendar_visibility = {
            id_: toggle.isChecked() for id_, toggle in self.calendar_toggles.items()
        }

        PREFERENCES.save()
        self.close()

    def update_calendars(self, calendars: list[Calendar]):
        """Display toggles for all calendar(s)"""

        num_items = len(self.calendar_toggles)
        self.calendar_toggles.clear()
        scroll_layout = self.scroll_widget.layout()

        # Clear the existing layout
        for i in reversed(range(scroll_layout.count())):
            widget = scroll_layout.itemAt(i).widget()
            if widget:
                widget.setParent(None)
            else:
                # Probably the spacer/stretch item
                item = scroll_layout.itemAt(i)
                scroll_layout.removeItem(item)

        # scroll_layout = QGridLayout()
        # scroll_layout.setContentsMargins(0, 0, 0, 0)
        for calendar in calendars:
            toggle = QCheckBox(
                f"Primary ({calendar.summary})"
                if calendar.primary
                else calendar.summary
            )
            if calendar.id not in PREFERENCES.calendar_visibility:
                visibility = True
            else:
                visibility = PREFERENCES.calendar_visibility[calendar.id]

            toggle.setChecked(visibility)
            toggle.setToolTip(f"ID: {calendar.id}")
            self.calendar_toggles[calendar.id] = toggle
            scroll_layout.addWidget(toggle)
        scroll_layout.addStretch()

        # Adjust the window size if calendar # has changed
        if num_items != len(self.calendar_toggles):
            self.adjustSize()



from __future__ import annotations

import datetime
import json
import os
import time
from ctypes import Union
from pprint import pprint as pp
from typing import Any, Optional

# Google imports
from google.auth.exceptions import RefreshError
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import Resource, build

from neverlate.utils import app_local_data_dir, now_datetime

# from googleapiclient.errors import HttpError

# If modifying these scopes, delete the file token.json.
SCOPES = ["https://www.googleapis.com/auth/calendar.readonly"]

EMAIL = None


class Calendar:
    """
    Google calendar data model
    """

    summary: str
    id: str
    primary: bool

    def __init__(self, data: dict[Any, Any]) -> None:
        self._data = data
        self.id = data["id"]
        self.primary = data.get("primary", False)
        self.summary = data.get("summaryOverride", data.get("summary", "<No Title>"))
        self.selected = data.get(
            "selected", False
        )  # Is the user displaying this calendar by default

    def __repr__(self):
        if self.primary:
            return "<Calendar: PRIMARY>"
        return (
            "<Calendar: "
            + " ".join(
                [
                    self.summary,
                    f"ID: {self.id}",
                    f"Primary: {self.primary}",
                ]
            )
            + ">"
        )


class TimeEvent:
    """
    Calendar event data model for time-based events (not all day events).
    """

    calendar: Calendar

    def __init__(self, item: dict[Any, Any], calendar: Calendar):
        """
        Create the time event.

        Args:
            item (dict): Dictionary result from event query.
            calendar (Calendar): Calendar this event belongs to.

        Raises:
            ValueError: If item is not valid/not a valid event with a start date and time.
        """
        if (
            "start" not in item
            or "dateTime" not in item["start"]
            or item.get("eventType", "default")
            != "default"  # or ['default', 'focusTime', 'outOfOffice']:
        ):
            raise ValueError("Invalid data type - not a calendar event")
        self.calendar = calendar
        self._event = item  # type: dict
        self.summary = self._event.get("summary", "<No Title>")

        # Get start and end times as datetime objects
        st_time = self._event["start"]["dateTime"]
        self.start_time = datetime.datetime.fromisoformat(st_time)
        # self.start_time = datetime.datetime.strptime(st_time[:19], "%Y-%m-%dT%H:%M:%S")
        end_time = self._event["end"]["dateTime"]
        self.end_time = datetime.datetime.fromisoformat(end_time)
        self.id = self._event["id"] + "::" + self.start_time.isoformat()

    def __repr__(self):
        return (
            "<TimeEvent: "
            + "   ".join(
                [
                    (
                        self.summary
                        if len(self.summary) <= 50
                        else self.summary[:47] + "..."
                    ).ljust(50),
                    "Seconds till event:  "
                    + f"{self.get_seconds_till_event()/60:.2f}".ljust(7),
                    # f"Event Type: {self._event['eventType']}",
                    # "End Time:", self._event['endTime'],
                    # f"ID: {self.id}",
                    f"{self.calendar.summary}",
                    "DECLINED" if self.has_declined() else "        ",
                    # f"Declined: {self.has_declined()}",
                ]
            )
            + ">"
        )

    def get_seconds_till_event(self) -> float:
        return (self.start_time - now_datetime()).total_seconds()

    def get_video_url(self) -> str:
        entry_points = self._event.get("conferenceData", {}).get("entryPoints", [])
        for entry_point in entry_points:
            if entry_point["entryPointType"] == "video":
                return entry_point["uri"]

        return ""

    def has_declined(self) -> bool:
        for attendee in self._event.get("attendees", []):
            # if attendee["email"] == "bwalters@wayfair.com":
            return attendee["responseStatus"] == "declined"
        return False

    def has_ended(self) -> bool:
        return (self.end_time - now_datetime()).total_seconds() < 0


class GoogleCalDownloader:  # (QObject):
    """Interface with Google. Credentials, downloading calendars, and downloading events."""

    primary_calendar: Calendar  # TODO: not used?
    calendars: list[Calendar] = []
    events: list[TimeEvent] = []
    last_update_time: float = 0.0

    # events_gathered_signal = Signal(list[Calendar])

    def __init__(self):
        super().__init__()
        self.service = None
        # self.service = build("calendar", "v3", credentials=self.creds)  # type: Resource
        self._cred_file_path = os.path.join(
            os.path.dirname(__file__), "credentials.json"
        )

        self.user_token_file_path = os.path.join(app_local_data_dir(), "token.json")

    def logout(self):
        if os.path.exists(self.user_token_file_path):
            os.remove(self.user_token_file_path)

        self.service = None

    def login(self) -> Credentials:
        """Gets google authentication credentails."""
        creds = self.get_existing_credentials()
        if not creds:
            flow = InstalledAppFlow.from_client_secrets_file(
                self._cred_file_path, SCOPES
            )
            creds = flow.run_local_server(port=0)

            # Save the credentials for the next run
            with open(self.user_token_file_path, "w") as token_file:
                token_file.write(creds.to_json())

        self.service = build("calendar", "v3", credentials=creds)  # type: Resource
        return creds

        creds = None
        # The file token.json stores the user's access and refresh tokens, and is
        # created automatically when the authorization flow completes for the first
        # time.
        if os.path.exists(self.user_token_file_path):
            creds = Credentials.from_authorized_user_file(
                self.user_token_file_path, SCOPES
            )

        # If there are no (valid) credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                try:
                    # TODO: sync with google's example code (and remove try/except ?)
                    creds.refresh(Request())
                except RefreshError:
                    print(" ================ Token is bad! ================")
                    try:
                        creds.refresh(Request())
                    except RefreshError:
                        os.remove(self.user_token_file_path)
                        return self.get_credentials()
            else:
                # Make the user sign in via a browser
                flow = InstalledAppFlow.from_client_secrets_file(
                    self._cred_file_path, SCOPES
                )
                creds = flow.run_local_server(port=0)

            # Save the credentials for the next run
            with open(self.user_token_file_path, "w") as token_file:
                token_file.write(creds.to_json())
        return creds

    def get_existing_credentials(self) -> Optional[Credentials]:
        """
        Check if we have valid credentials or not.

        Returns:
            Credentials|None
        """
        if not os.path.exists(self.user_token_file_path):
            return None
        creds = Credentials.from_authorized_user_file(self.user_token_file_path, SCOPES)
        if creds.valid:
            return creds
        if creds.expired and creds.refresh_token:
            try:
                creds.refresh(Request())

                # Save the credentials for the next run
                with open(self.user_token_file_path, "w") as token_file:
                    token_file.write(creds.to_json())

                return creds
            except RefreshError:
                pass

    def update_calendars(self) -> None:
        result = self.service.calendarList().list().execute()
        cal_ids = []
        for cal in result["items"]:
            if 0:
                print("=" * 80)
                pp(cal)
                continue
            if cal.get("deleted"):
                continue
            cal = Calendar(cal)
            cal_ids.append(cal)

        self.calendars = cal_ids

    def update_events(self, calendars: Optional[list[Calendar]] = None) -> None:
        events = []
        if calendars is None:
            calendars = self.calendars

        for calendar in calendars:
            events += self.get_events(calendar)
        self.events = events
        self.last_update_time = time.time()
        # self.events_gathered_signal.emit()

    def get_events(self, calendar: Calendar) -> list[TimeEvent]:
        """Get amazing events.

        Args:
            calendar (Calendar): calendar to get events from.

        Returns:
            list[TimeEvent]
        """
        # Call the Calendar API

        # now_ = datetime.datetime.utcnow().isoformat() + "Z"  # 'Z' indicates UTC time
        # time_min = now.replace(hour=0, minute=0, second=0, microsecond=0)
        # time_max = now.replace(hour=23, minute=59, second=59, microsecond=0)
        now_ = now_datetime()
        time_min = now_.replace(hour=0, minute=0, second=0)
        time_max = now_.replace(hour=23, minute=59, second=59)
        # print("Cal:", calendar.summary, "---", calendar.id)
        events_query = self.service.events().list(  # type: ignore
            calendarId=calendar.id,
            maxAttendees=1,
            timeMin=time_min.isoformat(),
            timeMax=time_max.isoformat(),
            maxResults=250,
            singleEvents=True,
            # orderBy="startTime",
        )
        events_result = events_query.execute()

        # pp(events_result)
        # print("   Items:", len(events_result["items"]))

        items = events_result.get("items", [])
        result = []
        # Prints the start and name of the next 10 events
        for item in items:
            try:
                event = TimeEvent(item, calendar)
            except ValueError:
                #  print("Invalid event", item.get("summary", "<No Title>"))
                continue
            result.append(event)

        return result


if __name__ == "__main__":
    gcal = GoogleCalDownloader()
    gcal.update_calendars()
    gcal.update_events()



from __future__ import annotations

import typing
from datetime import timedelta

from PySide6.QtCore import QRect, Qt, QThread, QTimer, Signal, Slot  # QThreadPool
from PySide6.QtGui import QAction, QColor, QCursor, QDesktopServices, QFont, QWindow
from PySide6.QtWidgets import (  # pylint: disable=no-name-in-module
    QAbstractScrollArea,
    QDialog,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QPushButton,
    QVBoxLayout,
)

from neverlate.utils import get_icon, now_datetime, pretty_datetime


class LoginDialog(QDialog):
    """Main dialog to show general info."""

    def __init__(self) -> None:
        super().__init__()
        self.setWindowTitle("NeverLate: Login")
        self.setWindowIcon(get_icon("tray_icon.png"))
        # self.setWindowFlag(Qt.Dialog)
        self.setWindowModality(Qt.ApplicationModal)

        main_layout = QVBoxLayout()
        main_layout.addWidget(
            QLabel(
                "Welcome to NeverLate!\n"
                "To get started, you need to log in to Google.\n"
                "Pressing 'Login' will open up your browser asking you to log in."
            )
        )
        self.login_button = QPushButton("Login")
        self.quit_button = QPushButton("Quit")
        self.quit_button.pressed.connect(self.reject)
        self.login_button.pressed.connect(self.accept)

        layout = QHBoxLayout()
        layout.addStretch()
        layout.addWidget(self.login_button)
        layout.addWidget(self.quit_button)

        main_layout.addLayout(layout)

        self.setLayout(main_layout)



"""Main app entry point."""
# pylint: disable=no-name-in-module
from __future__ import annotations

import ctypes
import logging
import sys
import time
import traceback

from google.auth.exceptions import RefreshError
from PySide6.QtCore import QRect, Qt, QThread, QTimer, Slot
from PySide6.QtWidgets import QApplication, QMenu, QMessageBox, QSystemTrayIcon

from neverlate.constants import APP_NAME
from neverlate.event_alerter import EventAlerter
from neverlate.google_cal_downloader import GoogleCalDownloader
from neverlate.login_dialog import LoginDialog
from neverlate.main_dialog import MainDialog
from neverlate.preferences import PREFERENCES
from neverlate.preferences_dialog import PreferencesDialog
from neverlate.utils import get_icon, seconds_to_min_sec

# TODO: add a column for attending status, join button, reset time alert button
# TODO: support calendars
# TODO: support preferences
# TODO: make it prettier for mac osx dark theme (just handle/bypass OS themes altogether?)


logger = logging.getLogger("NeverLate")


class UpdateCalendar(QThread):
    """Thread to download google calendars + events."""

    def __init__(self, calendar: GoogleCalDownloader) -> None:
        super().__init__()
        self.gcal = calendar
        self.needs_login = False

    def run(self):
        """Main entry point.

        Raises:
            RefreshError: When unable to download the calendar events for some reason.
        """
        self.needs_login = False
        try:
            self.gcal.update_calendars()
            calendars_to_update = [
                cal
                for cal in self.gcal.calendars
                if PREFERENCES.calendar_visibility.get(cal.id, cal.selected)
            ]
            self.gcal.update_events(calendars=calendars_to_update)
        except RefreshError:
            logger.error("BAD THINGS HAVE HAPPENED AND NEED TO BE FIXED")
            self.needs_login = True
        except ConnectionResetError:
            logger.debug(
                "ConnectionResetError while trying to download calendars + events. (Will try again)"
            )
        except ConnectionAbortedError:
            logger.debug(
                "ConnectionAbortedError error while trying to download calendars + events. (Will try again)"
            )
        except:
            line = "=" * 80
            logger.error("%s\n%s\n%s", line, traceback.format_exc(), line)


class App:
    """Main Qt application."""

    app: QApplication
    tray: QSystemTrayIcon
    preferences_dialog: PreferencesDialog

    def __init__(self) -> None:
        # Create a Qt application
        self.app = QApplication(sys.argv)
        self.app.aboutToQuit.connect(self.quitting)
        self.app.setQuitOnLastWindowClosed(False)
        self.main_dialog = MainDialog()
        self.main_dialog.update_now_button.clicked.connect(self.on_update_now)

        # Size of initial window
        rect = QRect(0, 0, 600, 300)
        screen_geo = self.app.primaryScreen().geometry()
        rect.moveCenter(screen_geo.center())
        self.main_dialog.setGeometry(rect)

        self.preferences_dialog = PreferencesDialog()
        self.preferences_dialog.logout_button.pressed.connect(
            lambda: self.login(force=True)
        )
        self._setup_tray()

        # Log in & get google calendar events
        self.gcal = GoogleCalDownloader()

        self.login()
        # self.gcal.update_calendars()
        # Timer - runs in the main thread every 1 second

        self.my_timer = QTimer()
        self.my_timer.timeout.connect(self.update)
        self.my_timer.start(1 * 1000)  # 1 sec intervall

        self.update_calendar_thread = UpdateCalendar(self.gcal)
        self.update_calendar_thread.finished.connect(
            self.thread_download_calendar_finished
        )
        self.update_calendar_thread.started.connect(
            self.thread_download_calendar_started
        )
        self.update_calendar_thread.start()

        self.event_alerters = {}  # type: dict[str, EventAlerter]

    def _setup_tray(self) -> None:
        menu = QMenu()
        main_dialog_action = menu.addAction("Show Overview")
        main_dialog_action.triggered.connect(self.show_main_dialog)
        setting_action = menu.addAction("Show Preferences")
        setting_action.triggered.connect(self.show_preferences_dialog)
        exit_action = menu.addAction("Quit")
        exit_action.triggered.connect(self.app.exit)
        self.tray = QSystemTrayIcon()
        self.tray.activated.connect(self.tray_clicked)
        self.tray.setIcon(get_icon("tray_icon.png"))
        self.tray.setContextMenu(menu)
        self.tray.show()
        self.tray.setToolTip("Never late!")
        # self.tray.showMessage(
        #     "My Great Title",
        #     "You're late for an event!",
        #     get_icon("tray_icon.png"),
        # )

    def login(self, force: bool = False):
        """Log into Google!

        Args:
            force (bool, optional): Force the user to re-log in.. Defaults to False.
        """
        if force:
            self.gcal.logout()
        elif self.gcal.login():
            return
        login_dialog = LoginDialog()
        login_dialog.login_button.pressed.connect(self.main_dialog.hide)
        login_dialog.login_button.pressed.connect(self.gcal.login)
        result = login_dialog.exec()
        if result == 0:  # User quit
            self.app.quit()  # TODO: test when in threads
            return sys.exit()

    def on_update_now(self):
        """User manually requested the calanders be re-downloaded."""
        self.update_calendar_thread.start()
        self.update()

    def quitting(self) -> None:
        """Quitting the app. Make sure we terminate all threads first."""
        self.update_calendar_thread.finished.disconnect()
        self.update_calendar_thread.terminate()
        for event in self.event_alerters.values():
            event.close_pop_up()

    def run(self):
        """Start the application."""
        if hasattr(ctypes, "windll"):
            # Rename the process so we can get a better icon.
            myappid = f"bw.{APP_NAME.lower()}.1"  # arbitrary string
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
        # self.main_dialog.show()
        self.app.setWindowIcon(get_icon("tray_icon.png"))
        self.main_dialog.show()  # TODO: enable/disable? just first time you launch?
        # Enter Qt application main loop
        self.app.exec()

    def show_main_dialog(self):
        """Show's the main dialog - forcing it to be on top."""
        # TODO: make this work for the preferences dialog (?)
        if sys.platform == "darwin":
            if self.main_dialog.isVisible():
                self.main_dialog.close()

            self.main_dialog.setWindowFlags(Qt.WindowStaysOnTopHint)
            self.main_dialog.show()
            return

        # Else: Windows/Linux
        if self.main_dialog.isVisible():
            # Close the dialog and re-open. This way we don't force the user to switch desktops/workspaces.
            self.main_dialog.close()
        self.main_dialog.setWindowFlags(
            # Qt.Tool  # Tool makes it visible on all workspaces/desktops for Windows
            Qt.Dialog
        )  # TODO: test / research more https://doc.qt.io/qt-5/qt.html#WindowType-enum
        self.main_dialog.show()
        self.main_dialog.activateWindow()

    def show_preferences_dialog(self):
        """Show the preferences dialog. Make sure it has the latest calendars."""
        self.preferences_dialog.update_calendars(self.gcal.calendars)
        self.preferences_dialog.show()

    def thread_download_calendar_finished(self):
        """
        Called when the update thread is finished - all google calendars and events have been donwloaded.
        """
        # Check if there was an error
        if self.update_calendar_thread.needs_login:
            mb = QMessageBox()
            mb.setWindowTitle("NeverLate: Logged Out")
            mb.setIcon(QMessageBox.Critical)
            mb.setText("You have been logged out of Google. Please log back in.")
            mb.setDetailedText(
                "This happened because this app has not been verified by Google yet, which requires "
                "testers to re-login once a week."
            )
            mb.setStandardButtons(QMessageBox.Ok)
            mb.exec()
            self.login(force=True)
            self.update_calendar_thread.start()
            return

        # Update the GUI
        self.main_dialog.update_now_button.setEnabled(True)

        # Check if new calendars need to be saved
        save_prefs = False
        for cal in self.gcal.calendars:
            if cal.id not in PREFERENCES.calendar_visibility:
                PREFERENCES.calendar_visibility[cal.id] = cal.selected
                save_prefs = True

        if save_prefs:
            PREFERENCES.save()

        # Process new/old events (close pop-ups for deleted events)
        cur_event_ids = {event.id for event in self.gcal.events}
        for time_event in self.gcal.events:
            if time_event.id not in self.event_alerters:
                self.event_alerters[time_event.id] = EventAlerter(time_event)
            else:
                self.event_alerters[time_event.id].time_event = time_event

        for id_ in set(self.event_alerters) - cur_event_ids:
            self.event_alerters[id_].close_pop_up()
            del self.event_alerters[id_]

    def thread_download_calendar_started(self):
        """Called when the thread to download calendars + events is triggered. Updates the UI accordingly."""
        self.main_dialog.time_to_update_label.setText("Updating events...")
        self.main_dialog.update_now_button.setEnabled(False)

    @Slot()
    def tray_clicked(self, reason: QSystemTrayIcon.ActivationReason):
        """Callback when the user clicks on the tray icon in the task bar. Should show various options/show the
        main GUI.
        """
        if sys.platform == "darwin":
            if reason == QSystemTrayIcon.ActivationReason.Trigger:
                return  # Mac OSX always shows the menu on left click (Trigger)
            else:
                logger.debug("UNHANDLED TRAY CLICK: %s", reason)
        else:
            # Windows/Linux
            if reason in (
                QSystemTrayIcon.ActivationReason.Trigger,
                QSystemTrayIcon.ActivationReason.DoubleClick,
            ):
                # Force show the main dialog
                if self.main_dialog.isVisible():
                    self.main_dialog.close()
                else:
                    self.show_main_dialog()

    def update(self):
        """Main update thread that should be continuously running."""

        if (
            not self.update_calendar_thread.needs_login
            and self.update_calendar_thread.isFinished()
        ):
            time_to_update = (PREFERENCES.download_cal_freq * 60) - (
                time.time() - self.gcal.last_update_time
            )
            if time_to_update <= 0:
                self.update_calendar_thread.start()
            else:
                time_label = seconds_to_min_sec(int(time_to_update))
                self.main_dialog.time_to_update_label.setText(
                    f"Updating events in {time_label}"
                )

        display_events = []
        for event_alerter in self.event_alerters.values():
            if not PREFERENCES.calendar_visibility.get(
                event_alerter.time_event.calendar.id, False
            ):
                # Calendar is not enabled, close pop-ups and move on
                # TODO: broken?
                event_alerter.close_pop_up()
            else:
                event_alerter.update()
                display_events.append(event_alerter)

        self.main_dialog.update_table_with_events(display_events)
        # self.main_dialog.setSizePolicy(QSizePolicy.Expanding)


def run():
    """Console tool entry point/ when run as __main__"""
    logging.basicConfig(
        level=logging.WARNING,
        format="NeverLate: %(asctime)s %(levelname)s : %(message)s",
    )
    if "--verbose" in sys.argv or "-v" in sys.argv:
        logger.setLevel(logging.DEBUG)
        logger.debug("Verbose mode is enabled!")

    app = App()
    app.run()


if __name__ == "__main__":
    run()



import ast
import re

import setuptools

_version_re = re.compile(r"__version__\s+=\s+(.*)")
with open("neverlate/__init__.py", "rb") as f:
    _match = _version_re.search(f.read().decode("utf-8"))
    if _match is None:
        raise SystemExit("No version found")
    version = str(ast.literal_eval(_match.group(1)))

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="neverlate",
    version=version,
    author="Brian Walters",
    author_email="brianrwalters@gmail.com",
    description="In your face notifications you can't miss for Google Calendar Events.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/beewally/neverlate",
    project_urls={
        "Bug Tracker": "https://github.com/beewally/neverlate/issues",
    },
    install_requires=[
        "PySide6",
        "google-api-python-client",
        "google-auth-httplib2",
        "google-auth-oauthlib",
    ],
    packages=[
        "neverlate",
        # "google-api-python-client",
        # "google-auth-httplib2",
        # "google-auth-oauthlib",
    ],  # setuptools.find_packages()
    # package_dir={"": "src"},
    package_data={"": ["credentials.json", "images/*.png"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: POSIX",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Office/Business",
        "Topic :: Office/Business :: Scheduling",
    ],
    python_requires=">=3.7",
    entry_points={
        "console_scripts": [
            "neverlate=neverlate.main:run",
        ]
    },
)



from __future__ import annotations

import datetime
import os

from PySide6.QtGui import QIcon

from neverlate.constants import APP_NAME

# LOCAL_TIMEZONE = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
LOCAL_TIMEZONE = datetime.datetime.now().astimezone().tzinfo
ICON_CACHE = {}  # type: dict[str, QIcon]


def app_local_data_dir() -> str:
    """
    Application data directory. Where we can safe preferences, tokens, etc.

    Returns:
        str: Folder path
    """
    if "HOME" in os.environ:
        # Linux/Mac
        folder = os.path.join(os.environ["HOME"], APP_NAME)
    else:
        # Windows
        folder = os.path.join(os.environ["APPDATA"], APP_NAME)

    if not os.path.exists(folder):
        os.makedirs(folder)
    return folder


def get_icon(name: str) -> QIcon:
    """Get an icon."""
    icon_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "images")
    if not name.endswith(".png"):
        name += ".png"
    icon = ICON_CACHE.get(name)
    if not icon:
        file_path = os.path.join(icon_dir, name)
        icon = QIcon(file_path)
        ICON_CACHE[name] = icon
    return icon


def now_datetime() -> datetime.datetime:
    """Current time in the users local time zone."""
    return datetime.datetime.now(LOCAL_TIMEZONE)


def pretty_datetime(dt: datetime.datetime) -> str:
    """Converts a date time to a a pretty syntax, e.g.: '4:33 PM'."""
    label = dt.strftime("%I:%M %p")
    if label[0] == "0":
        label = label[1:]
    return label


def seconds_to_min_sec(seconds: int) -> str:
    """Convert seconds to min:sec (e.g. 105 -> '1:45'."""
    min_, secs = divmod(seconds, 60)
    secs = str(secs).zfill(2)
    return f"{min_}:{secs}"



"""Module for displaying alerts to the user."""
from __future__ import annotations

import logging

# pylint: disable=no-name-in-module
import os
import subprocess  # nosec
import sys
from datetime import datetime, timedelta
from typing import Optional

from PySide6.QtCore import QThread, Signal, Slot

from neverlate.constants import OUTPUT_DISMISS, OUTPUT_SNOOZE
from neverlate.google_cal_downloader import TimeEvent
from neverlate.preferences import PREFERENCES
from neverlate.utils import now_datetime

logger = logging.getLogger("NeverLate")


class EventAlerter:
    """Alert about a timed event."""

    _snooze_until_time: Optional[datetime]
    time_event: TimeEvent
    has_alerted: bool  # Trure if an alert has been displayed ever
    dismissed_alerts: bool  # True if the user has dismissed the alert dialog
    _alerter: PopUpAlerterThread  # Thread monitoring subprocess for the pop-up dialogs

    def __init__(self, time_event: TimeEvent) -> None:
        self.time_event = time_event
        self.has_alerted = False
        self.dismissed_alerts = False
        self._alerter = PopUpAlerterThread(self.time_event)
        self._alerter.dismissed_signal.connect(self._dismiss_alerts)
        self._alerter.snooze_signal.connect(self.snooze)
        self._snooze_until_time = None

    @Slot()
    def _dismiss_alerts(self):
        """
        No longer display alerts for this event.
        """
        self.dismissed_alerts = True

    def close_pop_up(self):
        """Close any pop-up dialog threads.  Call before terminating."""
        try:
            if self._alerter.isRunning():
                logger.debug("TERMINATING A POP UP: %s", self.time_event.summary)
                if sys.platform == "win32":
                    subprocess.call(
                        ["taskkill", "/F", "/T", "/PID", str(self._alerter.process.pid)]
                    )
                else:
                    self._alerter.process.kill()
                    self._alerter.process.terminate()
                self._alerter.terminate()
        except RuntimeError:
            pass

    def reset_alert(self):
        """
        Reset any alerts. Used by the user to get a once-dismissed pop-up to appear again.
        """
        self.close_pop_up()
        self.dismissed_alerts = False
        self.has_alerted = False

    @Slot(int)
    def snooze(self, seconds: int = 0) -> None:
        """
        Snooze the alert until now + [seconds].

        Args:
            seconds (int): How many seconds to snooze for.
        """
        self._snooze_until_time = now_datetime() + timedelta(seconds=seconds)

    def time_till_alert(self) -> int:
        """Seconds until a notification should appear.

        Returns:
            int: Seconds till event. Less than zero = do not alert.
        """
        if (
            self.dismissed_alerts
            or self.time_event.has_declined()
            or self.time_event.has_ended()
            or self._alerter.isRunning()  # Already alerted
        ):
            return -1
        if self.has_alerted and self._snooze_until_time:
            secs_till_alert = (self._snooze_until_time - now_datetime()).total_seconds()
            return max(0, int(secs_till_alert))
        padding = 0 if self.has_alerted else PREFERENCES.alert_padding * 60
        secs_till_alert = (self.time_event.start_time - now_datetime()).total_seconds()
        return max(0, int(secs_till_alert) - padding)

    def update(self):
        """General function that should be called regularly to see if it should display an alert."""
        secs_till = self.time_till_alert()
        if secs_till != 0:
            return
        # Display an alert
        self.has_alerted = True
        self._alerter.start()

    def will_alert(self) -> bool:
        """Certain events will not alert (the user isn't attending, the meeting has ended already,
        the user dismissed the dialog, etc.  This function returns 'True' if this meeting will
        not alert.
        """
        # fmt: off
        if (
            self.dismissed_alerts
            or self.time_event.has_declined()  
            or self.time_event.has_ended() 
            or self._alerter.isRunning()  # Already alerted
        ):
            return False
        # fmt: on
        return True


class PopUpAlerterThread(QThread):
    """Thread to monitor subprocess of an alert dialog."""

    APP_PATH = os.path.join(os.path.dirname(__file__), "pop_up_alert.py")
    process: subprocess.Popen
    dismissed_signal = Signal()
    snooze_signal = Signal(int)

    def __init__(self, time_event: TimeEvent) -> None:
        super().__init__()

        self.time_event = time_event

    def run(self):
        """Main function to spawn a dialog and wait for it to be closed."""
        cmd = ["python", self.APP_PATH]
        cmd += [
            self.time_event.summary,
            self.time_event.start_time.isoformat(),
            self.time_event.get_video_url(),
        ]
        #  If MacOSX, no shell. If Windows, use shell. Linux = ??.  Otherwise pop-ups don't work. Not too sure why.
        use_shell = sys.platform == "win32"
        self.process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=use_shell,  # nosec
        )
        result = self.process.wait()
        output = self.process.stdout.read().decode()  # type: str
        err = self.process.stderr.read().decode()
        if result != 0:
            # Something bad happened. Just alert aga
            logger.error(err)
            return
        else:
            if len(output.splitlines()) > 1:
                logger.debug("Closed event, output:", output)
            output = output.splitlines()[-1] if output else ""
            if output.startswith(OUTPUT_SNOOZE):
                snooze_time = int(output.split()[-1])
                self.snooze_signal.emit(snooze_time)
            elif output.startswith(OUTPUT_DISMISS):
                self.dismissed_signal.emit()
